"""Store common constants."""
ACTION_TIMEOUT = 60.0
CONF_DATA = "data"
CONF_ID = "id"
CONF_TRIGGER = "trigger"
CONFIG_DIR = "config_dir"
LOG_LEVEL = "log_level"

# events
BASE_EVENT = "base_event"
CAMACQ_START_EVENT = "camacq_start_event"
CAMACQ_STOP_EVENT = "camacq_stop_event"
COMMAND_EVENT = "command_event"
IMAGE_EVENT = "image_event"
START_COMMAND_EVENT = "start_command_event"
STOP_COMMAND_EVENT = "stop_command_event"
